@@ -26,6 +26,7 @@
 typedef void (*sighandler_t)(int);
 #include <sysexits.h>
 #include <sys/time.h>
+#include <pthread.h>
 
 #ifdef __APPLE__
 #include <PCSC/wintypes.h>
@@ -48,7 +49,7 @@ typedef enum
 	True = TRUE, False = FALSE
 } Boolean;
 
-#define TIMEOUT 1000	/* 1 second timeout */
+#define TIMEOUT 3600*1000	/* 1 hour timeout */
 
 
 /* command used to parse (on screen) the ATR */
@@ -98,7 +99,9 @@ static void usage(const char *pname)
 	printf("%s usage:\n\n\t%s [ -h | -V | -n | -r | -c | -s | -t secs ]\n\n", pname, pname);
 	printf("\t\t  -h : this help\n");
 	printf("\t\t  -V : print version number\n");
+#ifndef WIN32
 	printf("\t\t  -n : no ATR analysis\n");
+#endif
 	printf("\t\t  -r : only lists readers\n");
 	printf("\t\t  -c : only lists cards\n");
 	printf("\t\t  -s : stress mode\n");
@@ -118,6 +121,7 @@ const char *cub3 = "";
 const char *cpl = "";
 
 time_t start_time;
+_Atomic Boolean Interrupted = False;
 
 typedef struct
 {
@@ -133,6 +137,8 @@ typedef struct
 
 static options_t Options;
 
+SCARDCONTEXT hContext;
+
 static Boolean is_member(const char *  item, const char * list[])
 {
 	int i = 0;
@@ -170,14 +176,14 @@ static void initialize_terminal(void)
 	}
 	else
 	{
-		cub2 = "\033[2D";
-		cub3 = "\033[3D";
+		cub2 = "\033[2D"; /* move back 2 characters */
+		cub3 = "\033[3D"; /* move back 3 characters */
 		cpl = "\033[";
 	}
 }
 /* There should be no \033 beyond this line! */
 
-static Boolean reached_maxtime(void)
+static Boolean should_exit(void)
 {
 	if (Options.maxtime)
 	{
@@ -185,82 +191,99 @@ static Boolean reached_maxtime(void)
 		if (t - start_time > Options.maxtime)
 			return True;
 	}
+
+#ifdef WIN32
+	if (GetKeyState(VK_SHIFT) & 0x80)
+		return True;
+#endif
+
+	if (Interrupted)
+		return True;
+
 	return False;
 }
 
-Boolean spinning_interrupted = False;
-unsigned int spin_state = 0;
+typedef enum
+{
+	SpinDisabled = -2,
+	SpinStopped = -1,
+	SpinRunning = 0
+} SpinState_t;
+
+_Atomic SpinState_t spin_state = SpinStopped;
+
 static void spin_start(void)
 {
-	spin_state = 0;
+	spin_state = Options.verbose ? SpinRunning : SpinDisabled;
 }
 
-static void spin_update(void)
+static void spin_stop(void)
+{
+	spin_state = SpinStopped;
+
+	/* clean previous output */
+	printf("%s %s ", cub2, cub2);
+	fflush(stdout);
+}
+
+static void *spin_update(void *p)
 {
 	char patterns[] = {'-', '\\', '|', '/'};
-	char c = patterns[spin_state];
 
-	if (reached_maxtime())
-		spinning_interrupted = 1;
+	/* 100 ms wait */
+	struct timespec wait_time = {.tv_sec = 0, .tv_nsec = 1000*1000*100};
 
-	if (! Options.verbose)
-		return;
+	(void)p;
 
-	spin_state++;
-	if (spin_state >= sizeof patterns)
-		spin_state = 0;
-	printf("%s %c ", cub3, c);
-	fflush(stdout);
-}
-
-static void spin_suspend(void)
-{
-	if (! Options.verbose)
-		return;
-
-	printf("%s %s", cub2, cub2);
-	fflush(stdout);
-	if (spinning_interrupted)
+again:
+	/* wait untill spinning starts */
+	do
 	{
-		printf("\n");
-		exit(EX_OK);
-	}
-}
+		if (should_exit())
+		{
+			if (SpinDisabled == spin_state)
+			{
+				SCardCancel(hContext);
+				pthread_exit(NULL);
+			}
+		}
 
-static sighandler_t old_interrupt_signal_handler;
-#define EX_USER_INTERRUPT (1)
+		nanosleep(&wait_time, NULL);
+	} while (spin_state < 0);
+
+	do
+	{
+		char c = patterns[spin_state];
+
+		if (should_exit())
+		{
+			SCardCancel(hContext);
+			pthread_exit(NULL);
+		}
+
+		spin_state++;
+		if (spin_state >= (int)sizeof patterns)
+			spin_state = SpinRunning;
+		printf("%s %c ", cub3, c);
+		fflush(stdout);
+
+		nanosleep(&wait_time, NULL);
+	} while (spin_state >= SpinRunning);
+
+	goto again;
+
+	return NULL;
+}
 
 static void user_interrupt_signal_handler(int signal)
 {
-	if (spinning_interrupted)
-	{
-		/*
-		 * If the user interrupts twice, before the program exits,
-		 * then we call the old interrupt signal handler,
-		 * or by default we exit.
-		 */
-
-		if (old_interrupt_signal_handler == SIG_IGN)
-		{
-			return;
-		}
-
-		if (old_interrupt_signal_handler == SIG_DFL)
-		{
-			exit(EX_USER_INTERRUPT);
-		}
-
-		old_interrupt_signal_handler(signal);
-	}
-	else
-	{
-		spinning_interrupted = True;
-	}
+	(void)signal;
+	Interrupted = True;
 }
 
 static void initialize_signal_handlers(void)
 {
-	old_interrupt_signal_handler = signal(SIGINT, user_interrupt_signal_handler);
+	signal(SIGINT, user_interrupt_signal_handler);
 }
 
 
@@ -280,10 +303,11 @@ static void initialize_options(options_t *options, const char *pname)
 	options->only_list_cards = False;
 }
 
+#define OPTIONS_BASE "Vhvrcst:q"
 #ifdef WIN32
-#define OPTIONS "Vhvrcst:q"
+#define OPTIONS OPTIONS_BASE
 #else
-#define OPTIONS "Vhnvrcst:q"
+#define OPTIONS OPTIONS_BASE "n"
 #endif
 
 static int parse_options(int argc, char *argv[], options_t *options)
@@ -348,7 +372,7 @@ static int parse_options(int argc, char *argv[], options_t *options)
 	return EX_OK;
 }
 
-static LONG stress(SCARDCONTEXT hContext, const char *readerName)
+static LONG stress(SCARDCONTEXT hContext2, const char *readerName)
 {
 	LONG rv, ret_rv = SCARD_S_SUCCESS;
 	SCARDHANDLE hCard;
@@ -361,7 +385,7 @@ static LONG stress(SCARDCONTEXT hContext, const char *readerName)
 	}
 	else
 		printf("\n");
-	rv = SCardConnect(hContext, readerName, SCARD_SHARE_SHARED,
+	rv = SCardConnect(hContext2, readerName, SCARD_SHARE_SHARED,
 			SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1, &hCard, &dwActiveProtocol);
 	if (rv != SCARD_S_SUCCESS)
 	{
@@ -409,6 +433,9 @@ static LONG stress(SCARDCONTEXT hContext, const char *readerName)
 			ret_rv = rv;
 			break;
 		}
+
+		if (Interrupted)
+			exit(EX_OSERR);
 	}
 
 	gettimeofday(&time_end, NULL);
@@ -460,7 +487,6 @@ int main(int argc, char *argv[])
 #else
 	LONG rv;
 #endif
-	SCARDCONTEXT hContext;
 	SCARD_READERSTATE *rgReaderStates_t = NULL;
 	SCARD_READERSTATE rgReaderStates[1];
 	DWORD dwReaders = 0, dwReadersOld;
@@ -471,6 +497,7 @@ int main(int argc, char *argv[])
 	char atr[MAX_ATR_SIZE*3+1];	/* ATR in ASCII */
 	char atr_command[sizeof(atr)+sizeof(ATR_PARSER)+2+1];
 	int pnp = TRUE;
+	pthread_t spin_pthread;
 
 	start_time = time(NULL);
 	initialize_terminal();
@@ -516,6 +543,9 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	/* start spining thread */
+	rv = pthread_create(&spin_pthread, NULL, spin_update, NULL);
+
 get_readers:
 	/* free memory possibly allocated in a previous loop */
 	if (NULL != readers)
@@ -596,14 +626,9 @@ get_readers:
 			do
 			{
 				rv = SCardGetStatusChange(hContext, TIMEOUT, rgReaderStates, 1);
-				spin_update();
-#ifdef WIN32
-				if (GetKeyState(VK_SHIFT) & 0x80)
-					spinning_interrupted = TRUE;
-#endif
 			}
-			while ((SCARD_E_TIMEOUT == rv) && !spinning_interrupted);
-			spin_suspend();
+			while (SCARD_E_TIMEOUT == rv);
+			spin_stop();
 
 			test_rv("SCardGetStatusChange", rv, hContext);
 		}
@@ -611,15 +636,20 @@ get_readers:
 		{
 			rv = SCARD_S_SUCCESS;
 			spin_start();
-			while ((SCARD_S_SUCCESS == rv) && (dwReaders == dwReadersOld) && !spinning_interrupted)
+			while ((SCARD_S_SUCCESS == rv) && (dwReaders == dwReadersOld))
 			{
 				rv = SCardListReaders(hContext, NULL, NULL, &dwReaders);
 				if (SCARD_E_NO_READERS_AVAILABLE == rv)
 					rv = SCARD_S_SUCCESS;
 				sleep(1);
-				spin_update();
+
+				if (should_exit())
+				{
+					printf("\n");
+					exit(EX_OK);
+				}
 			}
-			spin_suspend();
+			spin_stop();
 		}
 		if (Options.verbose)
 		{
@@ -691,7 +721,10 @@ get_readers:
 	 * We only stop in case of an error
 	 */
 	rv = SCardGetStatusChange(hContext, TIMEOUT, rgReaderStates_t, nbReaders);
-	while (((rv == SCARD_S_SUCCESS) || (rv == SCARD_E_TIMEOUT)) && !spinning_interrupted)
+
+	spin_stop();
+
+	while ((rv == SCARD_S_SUCCESS) || (rv == SCARD_E_TIMEOUT))
 	{
 		time_t t;
 
@@ -711,7 +744,6 @@ get_readers:
 					SCARD_STATE_CHANGED)
 #endif
 			{
-				spin_suspend();
 				goto get_readers;
 			}
 		}
@@ -721,14 +753,12 @@ get_readers:
 			if ((SCardListReaders(hContext, NULL, NULL, &dwReaders)
 				== SCARD_S_SUCCESS) && (dwReaders != dwReadersOld))
 			{
-				spin_suspend();
 				goto get_readers;
 			}
 		}
 
 		if (rv != SCARD_E_TIMEOUT)
 		{
-			spin_suspend();
 			/* Timestamp the event as we get notified */
 			t = time(NULL);
 			printf("\n%s", ctime(&t));
@@ -761,8 +791,6 @@ get_readers:
 			 * above.
 			 */
 
-			spin_suspend();
-
 			/* Specify the current reader's number and name */
 			printf(" Reader %d: %s%s%s\n", current_reader,
 				magenta, rgReaderStates_t[current_reader].szReader,
@@ -861,7 +889,7 @@ get_readers:
 				do
 				{
 					rv = stress(hContext, rgReaderStates_t[current_reader].szReader);
-				} while (SCARD_S_SUCCESS == rv && ! spinning_interrupted);
+				} while (SCARD_S_SUCCESS == rv);
 
 				rgReaderStates_t[current_reader].dwCurrentState = SCARD_STATE_UNAWARE;
 			}
@@ -870,25 +898,27 @@ get_readers:
 		if (Options.only_list_cards)
 			break;
 
+		spin_start();
+
 		rv = SCardGetStatusChange(hContext, TIMEOUT, rgReaderStates_t,
 			nbReaders);
 
-		spin_update();
-
-#ifdef WIN32
-	if (GetKeyState(VK_SHIFT) & 0x80)
-		spinning_interrupted = TRUE;
-#endif
+		spin_stop();
 	} /* while */
 
-	spin_suspend();
-
 	/* A reader disappeared */
 	if (SCARD_E_UNKNOWN_READER == rv)
 		goto get_readers;
 
 	/* If we get out the loop, GetStatusChange() was unsuccessful */
-	test_rv("SCardGetStatusChange", rv, hContext);
+	if (rv != SCARD_E_CANCELLED)
+		test_rv("SCardGetStatusChange", rv, hContext);
+	else
+	{
+		printf("%s", cub3);
+		fflush(stdout);
+		pthread_join(spin_pthread, NULL);
+	}
 
 	/* We try to leave things as clean as possible */
 	rv = SCardReleaseContext(hContext);
